#include <stdio.h>
#include <string.h>
#include <math.h>
#ifdef __WATCOMC__
#include <env.h>
#endif
#include <stdlib.h>
#include <dos.h>

#include "brender.h"
#include "dosio.h"
#include "fmt.h"

#define S0 BR_SCALAR(0.0)
#define S1 BR_SCALAR(1.0)

#define MAX_ARG_SIZE 80

// globals in for a quick fix
br_pixelmap    *screen=NULL, *back_screen=NULL, *depth_buffer=NULL, *pLogo=NULL;
br_pixelmap    *palette=NULL;

// standard stuff

#define FALSE 0
#define TRUE 0xFFFFFFFF
#define PmFree(x) if(x)BrPixelmapFree(x)


br_pixelmap *Load_Pixelmap( br_pixelmap *pScreen, char *Pixelmap )
{
	br_pixelmap *pMap;
	char Pixelmap_File[255];

	strcpy( Pixelmap_File, Pixelmap );

	// Load selected intro page as backdrop based on high or low resolution

	if ( pScreen->width==320 )
                strcat( Pixelmap_File, "_lw.GIF" );
	else if ( pScreen->width == 640 )
                strcat( Pixelmap_File, "_hg.GIF" );
	else
	{
		BR_ERROR0("Invalid screen mode");
		return FALSE ;
	}

        pMap = BrFmtGIFLoad( Pixelmap_File, BR_PMT_RGBX_888 );

	if ( pMap == NULL )
	{
		char Message[255];
		sprintf( Message, "Couldn't load %s", Pixelmap_File );
		BR_ERROR0( Message );
		return FALSE ;
	}

	return( pMap );
}



int Change_Screen_Res()
{
	BrZbEnd();
	DOSGfxEnd();
	
        PmFree(back_screen);
        PmFree(depth_buffer);
	PmFree(pLogo);
	
        if ( screen->width == 320 )
	{
                PmFree(screen);
                screen = DOSGfxBegin("VESA,W:640,H:480,B:8");
	}
	else
	{
                PmFree(screen);
                screen = DOSGfxBegin("MCGA");
	}

        BrZbBegin( screen->type,BR_PMT_DEPTH_16);
        back_screen  = BrPixelmapMatch( screen, BR_PMMATCH_OFFSCREEN );
        depth_buffer = BrPixelmapMatch( screen, BR_PMMATCH_DEPTH_16);
        BrPixelmapFill( back_screen, 0x0 );
        BrPixelmapFill( depth_buffer, 0xffffffff );

        pLogo = Load_Pixelmap( screen, "logo" );

	DOSGfxPaletteSet(palette);

	return TRUE ;
}

int Show_Help_Screen( br_pixelmap *pScreen, char *Help_Screen )
{
	br_pixelmap *pIntro ;
	char Help_File[255] ;

	strcpy( Help_File, Help_Screen );


	// Load selected intro page as backdrop based on high or low resolution

	if ( pScreen->width==320 )
                strcat( Help_File, "_lw.GIF" );
	else if ( pScreen->width == 640 )
                strcat( Help_File, "_hg.GIF" );
	else
	{
		BR_ERROR0("Invalid screen mode");
		return FALSE ;
	}

        pIntro = BrFmtGIFLoad(Help_File,BR_PMT_RGBX_888);

	if ( pIntro == NULL )
	{
		char Message[255];
		sprintf( Message, "Couldn't load %s", Help_File );
		BR_ERROR0( Message );
		return FALSE ;
	}

	// Display help screen

	BrPixelmapCopy( pScreen, pIntro );

	// Wait for the user to hit a key - must turn DOSKey off because
	// it currently has no 'check for any key' function.

	DOSKeyEnd();
	getch();
	DOSKeyBegin();


	BrPixelmapFree(pIntro);

	return TRUE ;
}

//End of standard stuff

void            LoadPalette(void);
void            LoadShadeTable(void);
void            LoadTexture(void);
void            LoadMaterial(void);
void            LoadModel(void);
br_actor       *LoadActors(void);

void
LoadPalette()
{
	palette = BrPixelmapLoad("os2pal.pal");
	if (palette == NULL)
		BR_ERROR0("Unable to zload palette.");
	DOSGfxPaletteSet(palette);
}

void
LoadShadeTable()
{
	br_pixelmap    *shade_table;
	shade_table = BrPixelmapLoad("os2shade.tab");
	if (!shade_table)
		BR_ERROR0("Failed to load shade table.");
	BrTableAdd(shade_table);
}

void
LoadTexture()
{
	br_pixelmap    *tile, *mask;
	tile = BrPixelmapLoad("tile01.pix");
	mask = BrPixelmapLoad("mask.pix");
	if (!((int) mask | (int) tile))
		BR_ERROR0("Failed to load textures.");

	BrMapAdd(tile);
	BrMapAdd(mask);
}

void
LoadMaterial()
{
	br_material    *material[10];
	int             n;

	n = BrMaterialLoadMany("robot.mat", material, 10);
	if (!n)
		BR_ERROR0("Failed to load materials.");

	BrMaterialAddMany(material, n);
}

void
LoadModel()
{
	br_model       *model[40];
	int             n;

	n = BrModelLoadMany("robot.dat", model, 40);
	if (!n)
		BR_ERROR0("Failed to load any models");

	BrModelAddMany(model, n);
}

br_actor       *
LoadActors()
{
	br_actor       *root;

	root = BrActorLoad("robot.act");
	if (root == NULL) {
		BR_ERROR0("Failed to load actor file.");
	}
	BrMatrix34RotateX(&root->t.t.mat, BR_ANGLE_DEG(-90.0));
	return root;
}

#define MAXFRAMES 100
#define MAXTRANSFORMS 2000

int             total_frames;

struct ani_transform {
	br_actor       *actor;
	br_matrix34     mat;
};

struct ani_frame {
	struct ani_transform *transforms;
	int             ntransforms;
};

struct ani_frame frames[MAXFRAMES];
struct ani_transform transforms[MAXTRANSFORMS];
br_actor       *ani_root;

void
LoadVUE(char *file_name, br_actor * root)
{
	FILE           *file;
	int             f = -1, n, i;
	char            buff[1024];
	char            name[80];
	br_model       *model;
	float           v[4][3];
	struct ani_transform *t = transforms;

	file = fopen(file_name, "r");

	while (!feof(file)) {
		fgets(buff, 1024, file);

		/* Is it a frame specification line? */

		if (sscanf(buff, "frame %d", &i) > 0) {
			if (f > 0)
				frames[f].ntransforms = t - frames[f].transforms;
			f = i;
			frames[f].transforms = t;
		}
		/* Or a transform specification? */

		if (sscanf(buff,
		 "transform \"%[^\"]\" %f %f %f %f %f %f %f %f %f %f %f %f",
				name,
				&v[0][0], &v[0][1], &v[0][2],
				&v[1][0], &v[1][1], &v[1][2],
				&v[2][0], &v[2][1], &v[2][2],
				&v[3][0], &v[3][1], &v[3][2]) > 0
			) {
			int             k, l;

			/* Found a transform so convert it to scalars */

			for (k = 0; k < 4; k++)
				for (l = 0; l < 3; l++)
					t->mat.m[k][l] = BrFloatToScalar(v[k][l]);

			t->actor = BrActorSearch(root, name);

			if (t->actor != NULL)
				t++;
		}
	}

	fclose(file);

	if (f > 0)
		frames[f].ntransforms = t - frames[f].transforms;

	total_frames = f + 1;
}



main()
{
	br_actor       *root;
	br_actor       *actors;
	br_actor       *camera;
	br_actor       *light;
	float           default_hither = 50.0;
	float           default_yon = 10000.0;
	int             n, new_cm;

	/* Start BRender */

	BrBegin();

	root = BrActorAllocate(BR_ACTOR_NONE, NULL);
	screen = DOSGfxBegin(NULL);
	if (screen == NULL) {
		BR_ERROR("Can't open screen mode");
	}
	back_screen = BrPixelmapMatch(screen, BR_PMMATCH_OFFSCREEN);
	depth_buffer = BrPixelmapMatch(back_screen, BR_PMMATCH_DEPTH_16);
	BrZbBegin(back_screen->type, depth_buffer->type);

	pLogo=Load_Pixelmap(back_screen,"logo");
	LoadPalette();
	LoadShadeTable();
	LoadTexture();
	LoadMaterial();
	LoadModel();
	actors = LoadActors();
	BrActorAdd(root, actors);
	LoadVUE("robot.vue", actors);
	if (root == NULL) {
		BR_ERROR0("No actor loaded");
	}
	light = BrActorAllocate(BR_ACTOR_LIGHT, NULL);
	((br_light *) light->type_data)->type = BR_LIGHT_POINT;
	((br_light *) light->type_data)->colour = BR_COLOUR_RGB(255, 255, 255);
	BrActorAdd(root, light);
	BrLightEnable(light);
	camera = BrActorAllocate(BR_ACTOR_CAMERA, NULL);
	camera->identifier = "camera";
	((br_camera *) camera->type_data)->type = BR_CAMERA_PERSPECTIVE;
	((br_camera *) camera->type_data)->field_of_view = BR_ANGLE_DEG(55.0);
	((br_camera *) camera->type_data)->hither_z = BrFloatToScalar(default_hither);
	((br_camera *) camera->type_data)->yon_z = BrFloatToScalar(default_yon);
	((br_camera *) camera->type_data)->aspect = BR_SCALAR(1.6);
	camera->t.type = BR_TRANSFORM_MATRIX34;
	BrActorAdd(root, camera);

	DOSKeyBegin();
	DOSMouseBegin();

	if(!Show_Help_Screen(screen,"intro"))exit(0);

	{
		float           x = 0;
		float           y = 0;
		float           z = 0;
		long            dx = 0;
		long            dy = 0;
		unsigned long   buttons;
		int             flash = 0;
		int             current_frame = 0;
		int             frame_delta = 1;
		float           sens;
		br_matrix34     rotation;
		br_matrix34     delta_rotation;
		br_matrix34     centre;
		br_matrix34     look;
		BrMatrix34Identity(&rotation);
		BrMatrix34Identity(&delta_rotation);
		BrMatrix34Identity(&centre);
		BrMatrix34Identity(&look);

		for (;;) {

			/*
			 * Key controls
			 */
			if (DOSKeyTest(SC_ESC, 0, 0))
				break;

			if (DOSKeyTest(SC_Q, 0, 0))
				break;

			if (DOSKeyTest(SC_MINUS, 0, REPT_FIRST_DOWN))
				current_frame--;

			if (DOSKeyTest(SC_EQUALS, 0, REPT_FIRST_DOWN))
				current_frame++;

			if (DOSKeyTest(SC_COMMA, 0, 0))
				frame_delta = -1;

			if (DOSKeyTest(SC_DOT, 0, 0))
				frame_delta = 1;

			if (DOSKeyTest(SC_SLASH, 0, 0))
				frame_delta = 0;

			if (DOSKeyTest(SC_H, 0,REPT_FIRST_DOWN))
				Show_Help_Screen(screen,"help");

			if (DOSKeyTest(SC_F1, 0,REPT_FIRST_DOWN))
				Show_Help_Screen(screen,"help");

			if (DOSKeyTest(SC_S, 0,REPT_FIRST_DOWN))
				Change_Screen_Res();

			/*
			 * Play any VUE file
			 */
			if (total_frames != 0) {
				int             i;

				/*
				 * Clamp frame number to valid range
				 */
				if (current_frame < 0)
					current_frame = total_frames - 1;

				if (current_frame >= total_frames)
					current_frame = 0;

				/*
				 * Apply animation to world
				 */
				for (i = 0; i < frames[current_frame].ntransforms; i++)
					frames[current_frame].transforms[i].actor->t.t.mat =
						frames[current_frame].transforms[i].mat;

				/*
				 * Move to next frame
				 */
				current_frame += frame_delta;
			}
			BrMatrix34Copy (&camera->t.t.mat, &rotation);
			BrMatrix34PreTranslate(&camera->t.t.mat, BR_SCALAR(x),
							 BR_SCALAR(y), BR_ADD(BR_SCALAR(z),BR_SCALAR(1700.0)));
			BrMatrix34Post(&camera->t.t.mat, &centre);

			BrMatrix34Copy(&light->t.t.mat, &camera->t.t.mat);

			BrPixelmapCopy(back_screen,pLogo);
			BrPixelmapFill(depth_buffer, 0xffffffff);

			if (flash) {
				flash = 0;
			} else {
				BrZbSceneRender(root, camera, back_screen, depth_buffer);
			}

			dx = 0;
			dy = 0;
			DOSMouseRead(&dx, &dy, &buttons);

			sens = sqrt(x * x + y * y + z * z) + 10;

			if (buttons == 0) {
				BrMatrix34RollingBall(&delta_rotation, -dx, dy, 800);
				BrMatrix34Pre(&rotation, &delta_rotation);
			} else if (buttons & BR_MSM_BUTTONR) {
				z -= dy * sens / 300.0;
				BrMatrix34PreRotateZ(&rotation, BR_ANGLE_DEG(dx / 10.0));
			} else if (buttons & BR_MSM_BUTTONL) {
				x -= dx * sens / 200.0;
				y += dy * sens / 200.0;
			}
			BrPixelmapDoubleBuffer(screen, back_screen);
		}
	}
	DOSMouseEnd();
	DOSKeyEnd();


	DOSGfxEnd();
	BrZbEnd();
	BrEnd();
	return 0;
}
